
gaze_data_fname                 = 'gazeData.tsv'
frame_timestamps_fname          = 'frameTimestamps.tsv'
scene_camera_video_fname_stem   = 'worldCamera'
scene_camera_calibration_fname  = 'calibration.xml'
gaze_overlay_video_file         = 'gazeOverlay.mp4'
marker_pose_prefix              = 'markerPose_'
fixation_classification_prefix  = 'fixation_classification'
fixation_assignment_prefix      = 'fixation_assignment'
validation_offset_fname         = 'dataQuality.tsv'